from os.path import join, basename
from os import makedirs
import json

from .vc_stub import VCStub
from .vc_sample import VCSample
from .constants import GVCF_EXT


def clone_project(project, target_dir, uuid=False, ext='.gvcf'):
    """Download stub files from a project to local storage.
    
    Return a list of paths to stub files.
    """
    name = project.uuid if uuid else project.name
    root_dir = join(target_dir, name)
    result_dir, sample_dir = join(root_dir, 'results'), join(root_dir, 'samples')
    makedirs(result_dir, exist_ok=True)
    makedirs(sample_dir, exist_ok=True)
    stub_files = []
    for result in project.get_analysis_results():
        stub_files += clone_result(result, result_dir, uuid=uuid, ext=ext, project=True)
    for sample in project.get_samples():
        stub_files += clone_sample(sample, sample_dir, uuid=uuid, ext=ext)
    return stub_files


def clone_sample(sample, target_dir, uuid=False):
    """Download stub files from a sample to local storage.
    
    Return a list of paths to stub files.
    """
    name = sample.uuid if uuid else sample.name
    root_dir = join(target_dir, name)
    makedirs(root_dir, exist_ok=True)
    gsq_dir = join(root_dir, '.geoseeq')
    makedirs(gsq_dir, exist_ok=True)
    vcsample = VCSample(sample, '.')
    vcsample.save_settings(gsq_dir)
    stub_files = []
    for stub, path in vcsample.stubs_from_remote():
        stub_path = path + GVCF_EXT
        stub.save_to_file(stub_path)
        stub_files.append(stub_path)
    return stub_files


def clone_result(result, target_dir, uuid=False, ext='.gvcf', project=False):
    """Download stub files from a result to local storage.
    
    Return a list of paths to stub files.
    """
    name = result.uuid if uuid else result.module_name + '__' + result.replicate
    root_dir = join(target_dir, name)
    makedirs(root_dir, exist_ok=True)
    stub_files = []
    for field in result.get_fields():
        stub_files.append(clone_field(field, root_dir, uuid=uuid, ext=ext, project=project))
    return stub_files


def clone_field(field, target_dir, uuid=False, ext='.gvcf', project=False):
    """Download a stub files from a field to local storage.
    
    Return the path to the stub file.
    """
    name = field.uuid if uuid else field.name
    name += ext
    target_path = join(target_dir, name)
    stub = VCStub(field.brn, field.get_local_filename(), field.checksum())
    stub.save_to_file(target_path)
    return target_path



import logging

import click


from .copy import cli_copy
from .manage import cli_manage
from .download import cli_download
from .upload import cli_upload
from .user import cli_user
from .view import cli_view
from geoseeq.vc.cli import cli_vc

logger = logging.getLogger('geoseeq_api')
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('[%(levelname)s] %(name)s :: %(message)s'))
logger.addHandler(handler)


@click.group()
def main():
    pass

main.add_command(cli_download)
main.add_command(cli_upload)
main.add_command(cli_manage)
main.add_command(cli_view)

@main.command()
def version():
    """Print the version of the Geoseeq API being used."""
    click.echo('0.2.9')  # remember to update setup


@main.group('advanced')
def cli_advanced():
    """Advanced commands."""
    pass

cli_advanced.add_command(cli_copy)
cli_advanced.add_command(cli_user)

@cli_advanced.group('experimental')
def cli_experimental():
    """Experimental commands."""
    pass

cli_experimental.add_command(cli_vc)





from time import sleep

from .bioproject import BioProject
from .setup_logging import logger


def create_geoseeq_group_from_bioproj(org, bioproj_accession, sleep_time=2):
    """Create a GeoSeeq project from an NCBI bioproject. Return the GeoSeeq project."""
    bioproj = BioProject(bioproj_accession)
    logger.info(bioproj)
    grp = bioproj.geoseeq_obj(org)
    grp.idem()
    logger.info(grp)

    samples, ars, fields = [], [], []
    for biosample in bioproj.biosamples():
        logger.info(biosample)
        sample = biosample.geoseeq_obj(grp)
        logger.info(sample)
        samples.append(sample)
        sleep(sleep_time)
        for sra in biosample.sra():
            logger.info(sra)
            ar, field = sra.geoseeq_obj(sample)
            logger.info(ar)
            logger.info(field)
            ars.append(ar)
            fields.append(field)
            sleep(sleep_time)

    for sample in samples:
        sample.idem()

    for ar in ars:
        ar.idem()

    for field in fields:
        field.idem()

    return grp




from .blob_constructors import (
    sample_from_blob,
    sample_result_folder_from_blob,
    sample_result_file_from_blob,
)
import json


def bulk_create_samples(knex, samples):
    """Create multiple samples at once. Returns a list of created samples.
    
    Only returns samples which were newly created.
    If a sample already exists on the server, it will not be returned.
    """
    data = {"samples": [sample.get_post_data() for sample in samples]}
    # print(json.dumps(data, indent=4))
    result = knex.post("bulk_samples", json=data)
    created_samples = [
        sample_from_blob(knex, result_blob) for result_blob in result['samples'] if result_blob
    ]
    return created_samples


def bulk_create_sample_result_folders(knex, sample_results):
    """Create multiple sample results at once. Returns a list of created sample results.
    
    Only returns sample results which were newly created.
    If a sample result already exists on the server, it will not be returned.
    """
    data = {"sample_results": [sample_result.get_post_data() for sample_result in sample_results]}
    result = knex.post(
        "bulk_sample_results",
        json=data,
    )
    created_sample_result_folders = [
        sample_result_folder_from_blob(knex, result_blob) for result_blob in result['sample_results'] if result_blob
    ]
    return created_sample_result_folders


def bulk_create_sample_result_files(knex, sample_result_fields):
    """Create multiple sample result fields at once. Returns a list of created sample result fields.
    
    Only returns sample result fields which were newly created.
    If a sample result field already exists on the server, it will not be returned.    
    """
    result = knex.post(
        "bulk_sample_result_fields",
        json={"sample_result_fields": [sample_result_field.get_post_data() for sample_result_field in sample_result_fields]},
    )
    created_sample_result_files = [
        sample_result_file_from_blob(knex, result_blob) for result_blob in result['sample_result_fields'] if result_blob
    ]
    return created_sample_result_files


import click
from geoseeq import Organization
from geoseeq.knex import GeoseeqNotFoundError



def _get_org(knex, org_name, yes, create=True):
    """Return an organization, creating it if necessary and desired."""
    try:
        org = Organization(knex, org_name).get()
    except GeoseeqNotFoundError:
        if not create:
            raise
        if not yes:
            click.confirm(f'Organization "{org_name}" does not exist. Create it?', abort=True)
        org = Organization(knex, org_name).create()
    return org


def _get_org_and_proj(knex, org_name, project_name, yes, private, create=True):
    """Return an organization and project, creating them if necessary and desired."""
    org = _get_org(knex, org_name, yes, create=create)
    try:
        proj = org.sample_group(project_name).get()
    except GeoseeqNotFoundError:
        if not create:
            raise
        if not yes:
            click.confirm(f'Project "{project_name}" does not exist. Create it?', abort=True)
        proj = org.sample_group(project_name, is_public=not private).create()
    return org, proj


def _get_org_proj_and_sample(knex, org_name, project_name, sample_name, yes, private, create=True):
    org, proj = _get_org_and_proj(knex, org_name, project_name, yes, private, create=create)
    try:
        sample = proj.sample(sample_name).get()
    except GeoseeqNotFoundError:
        if not create:
            raise
        if not yes:
            click.confirm(f'Sample "{sample_name}" does not exist. Create it?', abort=True)
        sample = proj.sample(sample_name).create()
    return org, proj, sample


def _get_org_proj_sample_and_folder(knex, org_name, project_name, sample_name, folder_name, yes, private, create=True):
    org, proj, sample = _get_org_proj_and_sample(knex, org_name, project_name, sample_name, yes, private, create=create)
    try:
        folder = sample.result_folder(folder_name).get()
    except GeoseeqNotFoundError:
        if not create:
            raise
        if not yes:
            click.confirm(f'Folder "{folder_name}" does not exist. Create it?', abort=True)
        folder = sample.folder(folder_name).create()
    return org, proj, sample, folder


def _get_org_proj_and_folder(knex, org_name, project_name, folder_name, yes, private, create=True):
    org, proj = _get_org_and_proj(knex, org_name, project_name, yes, private, create=create)
    try:
        folder = proj.result_folder(folder_name).get()
    except GeoseeqNotFoundError:
        if not create:
            raise
        if not yes:
            click.confirm(f'Folder "{folder_name}" does not exist. Create it?', abort=True)
        folder = proj.folder(folder_name).create()
    return org, proj, folder


import click
from requests.exceptions import HTTPError


def group_paired_end_paths(file_list, ext_1, ext_2, delim=''):
    """Return a dict mapping sample names -> "read_[1|2]" -> filepath.
    
    Raise a ValueError if a sample does not have exactly 2 reads.

    TODO: support multi lane files.
    """
    samples = {}
    for filepath in (l.strip() for l in file_list):
        if ext_1 in filepath:
            sname = filepath.split(ext_1)[0]
            key = 'read_1'
        elif ext_2 in filepath:
            sname = filepath.split(ext_2)[0]
            key = 'read_2'
        else:
            continue
        sname = sname.split('/')[-1]
        if delim:
            sname = sname.split(delim)[0]
        if sname not in samples:
            samples[sname] = {}
        samples[sname][key] = filepath
    for sname, reads in samples.items():
        if len(reads) != 2:
            raise ValueError(f'Sample {sname} has wrong number of reads: {reads}')
    return samples


def upload_fastq_pair(ar, read_1, read_2, private, overwrite=False):
    """Return two result fields with uploaded fastq files.
    
    No-op if the result fields already contained data and overwrite was not set.

    TODO: support multi lane files.
    """
    try:
        if overwrite:  # don't bother checking if the fields exist in overwrite mode
            raise HTTPError()
        r1 = ar.field('short_read::paired_end::read_1::lane_1').get()
        r2 = ar.field('short_read::paired_end::read_2::lane_1').get()
    except HTTPError:
        ar.is_private = private
        r1 = ar.field('short_read::paired_end::read_1::lane_1').idem()\
            .upload_file(
                read_1,
                optional_fields={
                    'format': 'fastq',
                    'sequencing_type': 'short_read::paired_end::read_1::lane_1',
                    'format_schema_version': 'v1',
                },
                logger=lambda x: click.echo(x, err=True)
            )
        r2 = ar.field('short_read::paired_end::read_2::lane_1').idem()\
            .upload_file(
                read_2,
                optional_fields={
                    'format': 'fastq',
                    'sequencing_type': 'short_read::paired_end::read_2::lane_1',
                    'format_schema_version': 'v1',
                },
                logger=lambda x: click.echo(x, err=True)
            )
    return r1, r2


def upload_single_fastq(ar, sequencing_type, filepath, private, overwrite):
    try:
        if overwrite:
            raise HTTPError()
        reads = ar.field(f'{sequencing_type}::read_1::lane_1').get()
    except HTTPError:
        ar.is_private = private
        reads = ar.field(f'{sequencing_type}::read_1::lane_1').idem()
        reads.upload_file(
            filepath,
            optional_fields={
                'format': 'fastq',
                'sequencing_type': f'{sequencing_type}::read_1::lane_1',
                'format_schema_version': 'v1',
            },
            logger=lambda x: click.echo(x, err=True),
        ) 
    return reads



from tqdm import tqdm
from os.path import basename

class TQBar:

    def __init__(self, pos, desc) -> None:
        self.n_bars = 0
        self.pos = pos
        self.desc = desc
        self.bar = None

    def set_num_chunks(self, n_chunks):
        self.n_bars = n_chunks
        self.bar = tqdm(total=n_chunks, position=self.pos, desc=self.desc, leave=False)

    def update(self, chunk_num):
        self.bar.update(chunk_num)


class PBarManager:

    def __init__(self):
        self.n_bars = 0
        self.pbars = []

    def get_new_bar(self, filepath):
        self.n_bars += 1
        return TQBar(self.n_bars, basename(filepath))


"""Test suite for experimental functions."""
import random
from os import environ
from os.path import dirname, join
from unittest import TestCase, skip

from geoseeq import Knex, Organization, RemoteObjectError, SampleGroup
from requests.exceptions import HTTPError

PACKET_DIR = join(dirname(__file__), "built_packet")
ENDPOINT = environ.get("GEOSEEQ_API_TESTING_ENDPOINT", "http://127.0.0.1:8000")


def random_str(len=12):
    """Return a random alphanumeric string of length `len`."""
    out = random.choices("abcdefghijklmnopqrtuvwxyzABCDEFGHIJKLMNOPQRTUVWXYZ0123456789", k=len)
    return "".join(out)


class TestGeoseeqApiClient(TestCase):
    """Test suite for packet building."""

    def setUp(self):
        self.knex = Knex(ENDPOINT)
        # Creates a test user and an API token for the user in database. Returns the token.
        api_token = self.knex.post("/users/test-user", json={"email": f"clitestuser_{random_str()}@gmail.com"})
        self.knex.add_api_token(api_token)

    def test_create_org(self):
        """Test that we can create an org."""
        org = Organization(self.knex, f"my_client_test_org {random_str()}")
        org.create()
        self.assertTrue(org.uuid)

    def test_get_org(self):
        """Test that we can get an org."""
        name = f"my_client_test_org {random_str()}"
        Organization(self.knex, name).create()
        org = Organization(self.knex, name).get()
        self.assertTrue(org.uuid)

    def test_idem_create_org(self):
        """Test that we can create an org using idem."""
        org = Organization(self.knex, f"my_client_test_org {random_str()}")
        org.idem()
        self.assertTrue(org.uuid)

    def test_idem_get_org(self):
        """Test that we can get an org using idem."""
        name = f"my_client_test_org {random_str()}"
        Organization(self.knex, name).create()
        org = Organization(self.knex, name).idem()
        self.assertTrue(org.uuid)

    def test_create_group(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        # N.B. It should NOT be necessary to call org.create()
        grp = org.sample_group(f"my_client_test_grp {key}")
        grp.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)

    def test_create_library(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}").create()
        grp_name = f"my_client_test_grp {key}"
        org.sample_group(grp_name, is_library=True).create()
        grp = SampleGroup(self.knex, org, grp_name).get()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(grp.is_library)

    def test_create_group_result(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        ar = grp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        ar.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(ar.uuid)

    def test_create_group_result_field(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}")
        ar = grp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(ar.uuid)
        self.assertTrue(field.uuid)

    def test_create_sample(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(samp.uuid)

    def test_add_sample(self):
        """Test that we can add a sample to a (non-library) sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        lib = org.sample_group(f"my_client_test_lib {key}", is_library=True)
        samp = lib.sample(f"my_client_test_sample {key}").create()

        grp = org.sample_group(f"my_client_test_grp {key}", is_library=False).create()
        grp.add_sample(samp).save()
        self.assertIn(samp.uuid, {samp.uuid for samp in grp.get_samples()})

    def test_get_samples_in_group(self):
        """Test that we can get the samples in a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp_names = [f"my_client_test_sample_{i} {key}" for i in range(10)]
        for samp_name in samp_names:
            grp.sample(samp_name).create()
        retrieved = org.sample_group(f"my_client_test_grp {key}", is_library=True).get()
        retrieved_names = set()
        for samp in retrieved.get_samples():
            retrieved_names.add(samp.name)
            self.assertTrue(samp.uuid)
        for samp_name in samp_names:
            self.assertIn(samp_name, retrieved_names)

    def test_get_results_in_group(self):
        """Test that we can get the results in a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        result_names = [("my_client_test_module", f"replicate_{i}") for i in range(10)]
        for module_name, replicate in result_names:
            grp.analysis_result(module_name, replicate=replicate).create()
        retrieved = org.sample_group(f"my_client_test_grp {key}", is_library=True).get()
        retrieved_names = set()
        for result in retrieved.get_analysis_results():
            retrieved_names.add((result.module_name, result.replicate))
            self.assertTrue(result.uuid)
        for result_name_rep in result_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_get_results_in_sample(self):
        """Test that we can get the results in a sample."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}").create()
        result_names = [("my_client_test_module", f"replicate_{i}") for i in range(10)]
        for module_name, replicate in result_names:
            samp.analysis_result(module_name, replicate=replicate).create()
        retrieved = grp.sample(f"my_client_test_sample {key}").get()
        retrieved_names = set()
        for result in retrieved.get_analysis_results():
            retrieved_names.add((result.module_name, result.replicate))
            self.assertTrue(result.uuid)
        for result_name_rep in result_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_get_result_fields(self):
        """Test that we can get the fields of an analysis result."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result("my_client_test_module").create()
        self.assertTrue(grp.uuid)

        field_names = [f"field_{i}" for i in range(10)]
        for field_name in field_names:
            ar.field(field_name).create()

        retrieved = samp.analysis_result("my_client_test_module").get()
        retrieved_names = set()
        for result in retrieved.get_fields():
            retrieved_names.add(result.name)
            self.assertTrue(result.uuid)
        for result_name_rep in field_names:
            self.assertIn(result_name_rep, retrieved_names)

    @skip("nonfunctional currently")
    def test_delete_sample(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(samp.uuid)
        samp.delete()
        self.assertRaises(lambda: setattr(samp, "name", "foo"), RemoteObjectError)
        retrieved = grp.sample(f"my_client_test_sample {key}")
        self.assertRaises(retrieved.get, HTTPError)

    def test_modify_sample_save(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(grp.is_public)
        self.assertTrue(samp.uuid)
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        samp.metadata = {f"metadata_{key}": "some_new_metadata"}
        self.assertTrue(samp._modified)
        samp.save()
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        retrieved = grp.sample(f"my_client_test_sample {key}").get()
        self.assertIn(f"metadata_{key}", retrieved.metadata)

    def test_modify_sample_idem(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(samp.uuid)
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        samp.metadata = {f"metadata_{key}": "some_new_metadata"}
        self.assertTrue(samp._modified)
        samp.idem()
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        retrieved = grp.sample(f"my_client_test_sample {key}").get()
        self.assertIn(f"metadata_{key}", retrieved.metadata)

    def test_create_sample_result(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        ar.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(samp.uuid)
        self.assertTrue(ar.uuid)

    def test_create_sample_result_field(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(samp.uuid)
        self.assertTrue(ar.uuid)
        self.assertTrue(field.uuid)

    def test_modify_sample_result_field_idem(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        field = ar.field(f"my_client_test_field_name {key}", {"foo": "bar"})
        field.create()
        self.assertTrue(field.uuid)
        field.stored_data = {"foo": "bizz"}  # TODO: handle deep modifications
        field.idem()
        retrieved = ar.field(f"my_client_test_field_name {key}").get()
        self.assertEqual(retrieved.stored_data["foo"], "bizz")

    def test_get_sample_group_manifest(self):
        """Test that we can get a group manifest."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        manifest = grp.get_manifest()
        self.assertTrue(manifest)

    def test_get_sample_manifest(self):
        """Test that we can get a group manifest."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        manifest = samp.get_manifest()
        self.assertTrue(manifest)



import json
import logging
import os
import time
import urllib.request
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile

import requests

from geoseeq.constants import FIVE_MB
from geoseeq.remote_object import RemoteObject, RemoteObjectError
from geoseeq.utils import download_ftp, md5_checksum

from .utils import *
from .file_upload import ResultFileUpload
from .file_download import ResultFileDownload


class ResultFile(RemoteObject, ResultFileUpload, ResultFileDownload):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "stored_data",
        "pipeline_run",
    ]
    optional_remote_fields = [
        "pipeline_run",
    ]
    parent_field = "parent"

    def __init__(self, knex, parent, field_name, pipeline_run=None, data={}):
        super().__init__(self)
        self.knex = knex
        self.parent = parent
        self.name = field_name
        self.stored_data = data
        self.pipeline_run = pipeline_run
        self._cached_filename = None  # Used if the field points to S3, FTP, etc
        self._temp_filename = False

    @property
    def is_sample_result(self):
        return isinstance(self, SampleAnalysisResultField)

    @property
    def brn(self):
        obj_type = "sample" if self.canon_url() == "sample_ar_fields" else "project"
        brn = f"brn:{self.knex.instance_code()}:{obj_type}_result_field:{self.uuid}"

    def nested_url(self):
        return self.parent.nested_url() + f"/fields/{self.name}"

    def get_blob_filename(self):
        sname = self.parent.parent.name.replace(".", "-")
        mname = self.parent.module_name.replace(".", "-")
        fname = self.name.replace(".", "-")
        filename = join(self.parent.parent.name, f"{sname}.{mname}.{fname}.json").replace(
            "::", "__"
        )
        return filename

    def get_referenced_filename_ext(self):
        try:
            key = [k for k in ["filename", "uri", "url"] if k in self.stored_data][0]
        except IndexError:
            raise TypeError("Cannot make a reference filename for a BLOB type result field.")
        ext = self.stored_data[key].split(".")[-1]
        if ext in ["gz"]:
            ext = self.stored_data[key].split(".")[-2] + "." + ext
        return ext

    def get_referenced_filename(self):
        ext = self.get_referenced_filename_ext()
        sname = self.parent.parent.name.replace(".", "-")
        mname = self.parent.module_name.replace(".", "-")
        fname = self.name.replace(".", "-")
        filename = join(self.parent.parent.name, f"{sname}.{mname}.{fname}.{ext}").replace(
            "::", "__"
        )
        return filename

    def get_local_filename(self):
        """Return a filename that can be used to store this field locally."""
        try:
            return basename(self.get_referenced_filename())
        except TypeError:
            return basename(self.get_blob_filename())

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["analysis_result"] = self.parent.uuid
        url = f"{self.canon_url()}/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.parent.idem()
        blob = self.knex.get(self.nested_url())
        self.load_blob(blob, allow_overwrite=allow_overwrite)

    def get_post_data(self):
        """Return a dict that can be used to POST this field to the server."""
        data = {
            "analysis_result": self.parent.uuid,
            "name": self.name,
            "stored_data": self.stored_data,
            "pipeline_run": self.pipeline_run,
        }
        return data

    def link_file(self, link_type, *args, **kwargs):
        if link_type == "s3":
            return self.link_s3(*args, **kwargs)
        elif link_type == "ftp":
            return self.link_ftp(*args, **kwargs)
        elif link_type == "sra":
            return self.link_sra(*args, **kwargs)
        elif link_type == "azure":
            return self.link_azure(*args, **kwargs)

    def link_s3(self, url, endpoint_url=None):
        """Link this field to an S3 object.

        Args:
            url (str): The URL of the S3 object.
            endpoint_url (str): The URL of the S3 endpoint.
        """
        if endpoint_url is None:
            if not url.startswith("https://"):
                raise ValueError("endpoint_url must be specified for non-HTTPS URLs.")
            endpoint_url = "https://" + url.split("https://")[1].split("/")[0]
            url = "s3://" + url.split(endpoint_url)[1][1:]
        self.stored_data = {
            "__type__": "s3",
            "uri": url,
            "endpoint_url": endpoint_url,
        }
        return self.save()

    def link_ftp(self, url):
        """Link this field to an FTP object.

        Args:
            url (str): The URL of the FTP object.
        """
        self.stored_data = {
            "__type__": "ftp",
            "uri": url,
        }
        return self.save()

    def link_sra(self, url):
        """Link this field to an SRA object.

        Args:
            url (str): The URL of the SRA object.
        """
        self.stored_data = {
            "__type__": "sra",
            "uri": url,
        }
        return self.save()

    def link_azure(self, url):
        """Link this field to an Azure object.

        Args:
            url (str): The URL of the Azure object.
        """
        endpoint_url = "https://" + url.split("https://")[1].split("/")[0]
        self.stored_data = {
            "__type__": "azure",
            "uri": url,
            "endpoint_url": endpoint_url,
        }
        return self.save()

    def _create(self):
        check_json_serialization(self.stored_data)
        self.parent.idem()
        data = self.get_post_data()
        blob = self.knex.post(f"{self.canon_url()}?format=json", json=data)
        self.load_blob(blob)

    def __del__(self):
        if self._temp_filename and self._cached_filename:
            os.remove(self._cached_filename)

    def pre_hash(self):
        return self.name + self.parent.pre_hash()

    def copy(self, new_parent, save=True):
        copied = new_parent.field(self.name, data=self.stored_data)
        if save:
            copied.idem()
        return copied

    def checksum(self):
        """Return a checksum for this field as a blob.

        TODO
        """
        return {"value": "", "method": "none"}

AnalysisResultField = ResultFile

class SampleResultFile(ResultFile):
    def canon_url(self):
        return "sample_ar_fields"

    def __str__(self):
        return f"<Geoseeq::SampleResultFile {self.name} {self.uuid} />"

SampleAnalysisResultField = SampleResultFile


class ProjectResultFile(ResultFile):
    def canon_url(self):
        return "sample_group_ar_fields"

    def __str__(self):
        return f"<Geoseeq::ProjectResultFile {self.name} {self.uuid} />"

SampleGroupAnalysisResultField = ProjectResultFile


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools

setuptools.setup(
    name='geoseeq',
    version='0.2.9',  # remember to update version string in CLI as well
    author="David C. Danko",
    author_email='dcdanko@biotia.io',
    description=open('README.md').read(),
    packages=setuptools.find_packages(),
    package_dir={'geoseeq': 'geoseeq'},
    install_requires=[
        'requests',
        'click',
        'pandas',
        'biopython',
        'tqdm',
    ],
    entry_points={
        'console_scripts': [
            'geoseeq=geoseeq.cli:main'
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
)



import json

import click

from .utils import use_common_state


@click.group('user')
def cli_user():
    """Manage users on GeoSeeq."""
    pass


@cli_user.command('info', deprecated=True)
@use_common_state
def cli_user_info(state):
    knex = state.get_knex()
    response = knex.get('/users/me')
    click.echo(json.dumps(response, indent=4, sort_keys=True))


# New endpoint has to be created on server and can be done with clerk backend API

# @cli_user.command('change-password')
# @use_common_state
# @click.argument('new_password')
# def cli_change_password(state, new_password):
#     """Change a users password."""
#     knex = state.get_knex()
#     response = knex.post(
#         '/auth/users/set_password',
#         json={
#             'new_password': new_password,
#             'current_password': state.password,
#         }
#     )
#     click.echo(f'changed password for user: {state.email}', err=True)
